/*
   CCD defects reader
   by paul69, v0.12, 2003-08-20, <e2500@narod.ru>

   Supported cameras:
      E2500 (tested, except black remapping)
      E995 (tested: reading the hot map)
      E4500 (tested: remap hot pixels)
*/
#define STRICT
#include <windows.h>
#include <commctrl.h>
#pragma hdrstop

#include <stdio.h> // sscanf();
#include "rc.h"
#include "dsc_connect.h"

#define program STR(PROGRAM_CAPTION)
#define version "v0.12"
#define compilation_time __DATE__ " " __TIME__

#define IDC_CONNECT 200
#define IDC_CHECK_CONNECTION 201

#define STR(id) (Language::string_table[Language::id][Language::curr_lang])
struct Language
{
enum { EN, RU, LanguageCount };
enum {MESSAGE_LANGUAGE,

   CONNECTING_TO_CAMERA,
   CAMERA_CONNECTED,

   STOP_WARNING,
   READ_WARNING,
   REMAP_WARNING,
   ELAPSED_TIME,

   REMAP_FAULT,

   REMAP_OKX,
   REMAP_OK1,
   REMAP_OK0,
   REMAP_COMPLETE,
   REMAP_INTERNAL_ERROR,
   TIME_IS_UP,

   CANNOT_OPEN_CAMERA,
   PROGRAM_CAPTION,

   BUTTON_CLOSE,
   BUTTON_STOP,
   BUTTON_READ_DEFECT,
   BUTTON_REMAP_DEFECT,
   BUTTON_OPTIONS,

   CHECKBOX_ENABLE_PARSING,
   STATIC_LANGUAGE,

   CAN_NOT_START_REMAP,
   REMAP_CCD_DEFECTS,
   READ_ELAPSED_TIME,
   READ_ELAPSED_TIME_RAW,
   NOT_IMPLEMENTED,
   READ_CCD_DEFECTS,

   MANUAL_CONNECTION,
   CHOOSE_AND_PRESS_CONTINUE,
   BUTTON_CANCEL,
   BUTTON_CONNECT,
   StringCount };

   static const char* string_table[StringCount][LanguageCount];
   static int curr_lang;
};

int Language::curr_lang = Language::EN;

const char* Language::string_table[Language::StringCount][Language::LanguageCount] = {
{  "English",
   "" },

{  "Connecting to camera...",
   "  ..." },
{  "Camera connected via %s",
   "   %s" },

{  "Stop reading? It may take several seconds.",
   " ?     ." },

{  "Reading of the defect pixel's map from your camera. \r\n"
   "It may take about 5 minutes and can be aborted at any time. \r\n"
   "You are using this capability on you own risk. \r\n"
   "If your camera will be damaged, you may be voiding \r\n"
   "the warranty of the camera.\r\n"
   "Continue?",

   "         . \r\n"
   "     ,       . \r\n"
   "       . \r\n"
   "      ,      . \r\n"
   "?"},

{  "The automatic remapping of the CCD defects will be made now. \r\n"
   "You camera will take a shot with exposure about 5 seconds\r\n"
   "and process it. \r\n"
   "You are using this capability on you own risk. \r\n"
   "If your camera will be damaged, you may be voiding \r\nthe warranty of the camera.\r\n"
   "Continue?",

   "        . \r\n"
   "        5    . \r\n"
   "       . \r\n"
   "      ,      . \r\n"
   "?"},

{  "Elapsed time: %u sec\r\n",
   " : %u sec\r\n"},
{  "Remap is failure or it is not supported\r\n",
   "      .\r\n"},

{  "Remap complete. (%u sec)\r\nIt's been found %u defect pixels.\r\n",
   " . (%u )\r\n %u  .\r\n"},
{  "Remap complete. (%u sec)\r\nIt's been found one defect pixel.\r\n",
   " . (%u )\r\n   .\r\n"},
{  "Remap complete. (%u sec)\r\nIt hasn't been found any defect pixels.\r\n",
   " . (%u )\r\n   .\r\n"},
{  "Remap complete. (%u sec)\r\nResult is undefined.\r\n",
   " . (%u )\r\n   .\r\n"},
{  "Remap complete. (%u sec)\r\nInternal firmware error occur.\r\n",
   " . (%u )\r\n    \r\n"},
{  "Expected time is over. (%u sec)\r\n",
   "   . (%u )\r\n"},

{  "Cannot connect to camera in an auto mode.\r\n"
   "Check USB or COM connection and try again.\r\n"
   "Would you like try a manual connection mode?",

   "     .\r\n"
   "   USB  COM    .\r\n"
   "   ?" },

{  "CCD defects reader " version,
   "   " version},

{  "Close",
   "" },
{  "Stop",
   "" },
{  "Read pixels",
   "" },
{  "Remap pixels",
   "" },
{  "Options",
   "" },

{  "Enable to camera output parsing",
   "   " },
{  "Message / warning language:",
   " :" },

{  "Cannot start remap procedure.\r\n",
   "   .\r\n"},
{  "Remapping the defect pixels.\r\n",
   "  .\r\n"},
{  "Elapsed time: %5u sec (%3u%%).\r\nRead %4u defect pixels.",
   " : %5u  (%3u%%).\r\n %4u  ."},
{  "Elapsed time: %5u sec (%3u%%).\r\nUnparsed camera responces:",
   " : %5u  (%3u%%).\r\n  :"},
{  "Not inplemented!\r\nTry again next time, please.",
   " !\r\n   ."},
{  "Reading the defect pixels.\r\n",
   "   .\r\n"},

{  "Manual connection",
   " "},

{  "Choose drive or port from device list and, if\r\n"
   "you are sure that device is Coolpix digital camera,\r\n"
   "press [Connect] to continue... \r\n"
   "Please, don't try connecting with another device!",

   "       , \r\n"
   "  ,     Coolpix,\r\n"
   "  []  ... \r\n"
   "     !" },

{  "Cancel",
   ""},
{  "Connect",
   ""},
};


DSC* dsc;
HINSTANCE hAppInst;
HWND hAppWin;
HWND hOptDlg;
HWND hManDlg;
HWND hMsgWin;
HWND hProgressWin;

void EnableControls(HWND hDlg, BOOL en);

char buffer[0x10000];

enum { IDT_REMAP = 1, IDT_READ, 

   REMAP_TICK = 500,
   MAX_REMAP = 60*1000/REMAP_TICK,

   READ_TICK = 1000,
   MAX_READ = 52,

   CLEAR_TICK = 500,
   MAX_CLEAR = 10*1000/CLEAR_TICK,
};

bool is_busy;
unsigned start;
char* command;
char* status;
char* message;

int remap_count;
int remap_status;
enum { REMAP_ERROR=-2, REMAP_IN_PROGRESS, REMAP_INTERNAL_ERROR, REMAP_COMPLETE, REMAP_OK };

DWORD __stdcall StartRemap( void* )
{
   DOUT(("Start remapping defect pixels\n"))
   remap_status = REMAP_IN_PROGRESS;

   dsc->SetStr(0x1F, "");
   dsc->SetReg(0x20, -1);

   if( !dsc->SpecCmd(6, 0, 15000) )
   {
      DOUT(("SpecCmd(6, 0x00) - fault\n"))
   }

   for(int i=0; i<45; ++i)
   {
      TResponce responce;
      if( dsc->GetVar(0x1F, responce) )
      {
         const char* data = responce.PrepareStr();

         DOUT(("GetVar(0x1F): %s\n", data))

         if( data[0] == 'd' )
         {
            DOUT(("Remap internal error\n"))
            remap_status = REMAP_INTERNAL_ERROR;
            return TRUE;
         }
      }
      else
      {
         DOUT(("dsc->GetVar(0x1F, responce) - fault\n"))
      }

      unsigned defect_count;
      if( dsc->GetReg(0x20, defect_count) )
      {
         DOUT(("GetVar(0x20): %d\n", defect_count))

         if( (int)defect_count >= 0 )
         {
            DOUT(("Remap complete\n"))
            remap_status = REMAP_OK + defect_count;
            return TRUE;
         }
         break;
      }
      else
      {
         DOUT(("dsc->GetReg(0x20, defect_count) - fault\n"))
      }
      
      Sleep(1000);
   }

   DOUT(("Remap complete\n"))
   remap_status = REMAP_COMPLETE;
   return TRUE;
}

bool Remap(HWND hDlg, int is_black)
{
   DOUT(("Remapping\n"))

   if( is_busy ) {
      DOUT(("Remap: busy\n"))
      return false;
   }
   if( IDYES != MessageBox( hDlg, STR(REMAP_WARNING), program, MB_YESNOCANCEL|MB_ICONQUESTION ) ) {
      DOUT(("Remap: Cancel remapping\n"))
      return false;
   }
   EnableControls( hDlg, FALSE );
   remap_count = 0;

   SendMessage(hProgressWin, PBM_SETRANGE, 0, MAKELPARAM(0, MAX_REMAP) );
   SendMessage(hProgressWin, PBM_SETPOS, 0, 0 );

   start = GetTickCount();

   DWORD id;
   HANDLE h = CreateThread(NULL, 0x10000, StartRemap, (void*)is_black, 0, &id);
   if( h == NULL )
   {
      DOUT(("Remap: CreateThread(), error=%u\n", GetLastError()))
      remap_status = REMAP_ERROR;
      strcpy(command, STR(CAN_NOT_START_REMAP));
      if( hMsgWin ) SetWindowText(hMsgWin, buffer);
   }
   else
   {
      CloseHandle( h );

      is_busy = true;
      remap_status = REMAP_IN_PROGRESS;

      strcpy(command, STR(REMAP_CCD_DEFECTS));
      if( hMsgWin ) SetWindowText(hMsgWin, buffer);

      status = command + strlen(command);
      SetTimer(hDlg, IDT_REMAP, REMAP_TICK, NULL);
   }
   return true;
}

int clear_count;
int clear_status;
enum { CLEAR_ERROR=-2, CLEAR_IN_PROGRESS, CLEAR_COMPLETE, CLEAR_OK };

DWORD __stdcall StartClear( void* )
{
   DOUT(("Start clearing\n"))
   clear_status = CLEAR_IN_PROGRESS;

   if( !dsc->SetReg(0x20, 0x49) )
   {
      DOUT(("SetReg(0x20, 0x49) - fault\n"))
   }

   if( !dsc->SpecCmd(6, 0x16, 3000) )
   {
      DOUT(("SpecCmd(6, 0x16) - fault\n"))
   }

   for(int i=0; i<30; ++i)
   {
      TResponce responce;

      if( dsc->GetVar(0x1F, responce) )
      {
         DOUT(("GetVar(0x1F): %s\n", responce.PrepareStr()))
         DOUT(("Clear OK\n"))
         clear_status = CLEAR_OK;
         return TRUE;
      }
      else
      {
         DOUT(("GetVar(0x1F) - fault\n"))
      }
      Sleep(1000);
   }

   DOUT(("Clear complete\n"))
   clear_status = CLEAR_COMPLETE;
   return TRUE;
}

int pixel_count;
int read_count;
int read_status;
bool break_flag;
bool pause_flag;
bool show_stop;

bool enable_parcing = true;

enum { READ_START = -3, READ_IN_PROGRESS, READ_ERROR, READ_OK };

void SetStatusLine()
{
   wsprintf( status, enable_parcing ? STR(READ_ELAPSED_TIME) : STR(READ_ELAPSED_TIME_RAW),
      (GetTickCount() - start + 500)/1000, read_count*100/MAX_READ, pixel_count);
}

DWORD __stdcall StartRead( void* )
{
   DOUT(("Start reading defect pixels\n"))
   int max_pixels = 512;
   int max_tries = 52;
   pixel_count = 0;
   int i, x, y;

   read_status = READ_IN_PROGRESS;

   int last_x, last_y;
   last_x = last_y = 0;

   for(read_count=0; read_count<max_tries; ++read_count)
   {
      while( pause_flag ) Sleep(500);
      if( break_flag ) {
         DOUT(("Read: break\n"))
         break;
      }

      if( !dsc->SetReg(0x20, read_count) )
      {
         DOUT(("Read: SetReg(0x20, %d) - fault\n", read_count))
      }

      if( !dsc->SpecCmd(6, 0x14, 3000) )
      {
         DOUT(("Read: SpecCmd(6, 0x14) - fault\n"))
         read_status = READ_ERROR;
         return TRUE;
      }

      TResponce responce;

      for(i=0; i<30; ++i)
      {
         if( dsc->GetVar(0x1F, responce) )
         {
            DOUT(("Read: GetVar(0x1F): %s\n", responce.PrepareStr()))
            break;
         }
         else
         {
            DOUT(("Read: GetVar(0x1F) - fault\n"))
         }
         Sleep(500);
      }
      if( i == 30 )
      {
         DOUT(("Read: exceed timeout\n"))
         read_status = READ_ERROR;
         return TRUE;
      }

      if( break_flag ) {
         DOUT(("Read: break\n"))
         break;
      }

      if( !dsc->SetReg(0x20, 3) )
      {
         DOUT(("Read: SetReg(0x20, 0x03) - fault\n"))
      }

      if( !dsc->SpecCmd(6, 0x16, 3000) )
      {
         DOUT(("Read: SpecCmd(6, 0x16) - fault\n"))
         read_status = READ_ERROR;
         return TRUE;
      }

      for(i=0; i<30; ++i)
      {
         if( dsc->GetVar(0x1F, responce) )
         {
            DOUT(("Read: GetVar(0x1F): %s\n", responce.PrepareStr()))
            break;
         }
         else
         {
            DOUT(("Read: GetVar(0x1F) - fault\n"))
         }
         Sleep(500);
      }
      if( i == 30 )
      {
         DOUT(("Read: exceed timeout\n"))
         read_status = READ_ERROR;
         return TRUE;
      }

      const char* data = responce.PrepareStr();

      int len = strlen(message+2)+2;

      if( !enable_parcing )
      {
         wsprintf( message+len, "%s\r\n", (const char*)data );
      }
      else
      {
         const char* end = strchr(data, '\n');
         if( !end )
         {
            wsprintf( message+len, "Unknown responce format:\r\n%s\r\n", data);
            break;
         }

         for(i=0; i<16; ++i)
         {
            const char* cur = end + 1;
            end = strchr(cur, '\n');
            if( !end ) break;

            x = y = -1;
            sscanf(cur, "%i,%i", &x, &y);

            if( x == -1 )
               break;

               if( i == 0 )
               {
                  if( read_count == 1)
                  {
                     if( last_x == x && last_y == y )
                        break;
                  }
                  last_x = x;
                  last_y = y;
               }

            len += wsprintf( message+len, "%d\t%d\t%d\r\n", pixel_count+1, x, y );

            if( ++pixel_count == max_pixels )
               break;
         }
         if( i<10 ) break;
      }
   }

   DOUT(("Reading finished\n"))
   read_status = READ_OK;
   return TRUE;
}

bool Stop(HWND hDlg)
{
   pause_flag = true;
   if( IDYES != MessageBox( hDlg, STR(STOP_WARNING), program, MB_YESNOCANCEL|MB_ICONQUESTION ) )
   {
      pause_flag = false;
      return false;
   }

   break_flag = true;
   pause_flag = false;

   return true;
}

bool Read(HWND hDlg, int is_black)
{
   DOUT(("Reading\n"))

   if( is_busy ) {
      DOUT(("Read: busy\n"))
      return false;
   }
   if( IDYES != MessageBox( hDlg, STR(READ_WARNING), program, MB_YESNOCANCEL|MB_ICONQUESTION ) ) {
      DOUT(("Read: busy\n"))
      return false;
   }
   EnableControls( hDlg, FALSE );
   EnableWindow( GetDlgItem(hDlg, IDCANCEL), TRUE );
   SetFocus( GetDlgItem(hDlg, IDCANCEL) );

   break_flag = false;
   pause_flag = false;
   read_count = 0;
   read_status = READ_START;
   SetTimer(hDlg, IDT_READ, READ_TICK, NULL);
   SendMessage(hProgressWin, PBM_SETRANGE, 0, MAKELPARAM(0, MAX_READ) );
   SendMessage(hProgressWin, PBM_SETPOS, 0, 0 );

   start = GetTickCount();

   strcpy(command, STR(READ_CCD_DEFECTS));
   status = command + strlen(command);

   SetStatusLine();
   message = command + strlen(command);

   message[0] = '\r';
   message[1] = '\n';
   message[2] = '\0';

   if( hMsgWin ) SetWindowText(hMsgWin, buffer);

   DWORD id;
   HANDLE h = CreateThread(NULL, 0x10000, StartRead, (void*)is_black, 0, &id);
   if( h == NULL )
   {
      DOUT(("Read: CreateThread(), error=%u\n", GetLastError()))
      read_status = READ_ERROR;
   }
   else
   {
      is_busy = true;
      CloseHandle( h );

      show_stop = true;
      SetDlgItemText(hDlg, IDCANCEL, STR(BUTTON_STOP));
   }
   return true;
}

bool SetLanguage( int new_lang )
{
   if( new_lang >= 0 && new_lang < Language::LanguageCount )
   {
      Language::curr_lang = new_lang;
      SetWindowText( hAppWin, STR(PROGRAM_CAPTION) );
      SetDlgItemText( hAppWin, IDCANCEL, STR(BUTTON_CLOSE) );
      SetDlgItemText( hAppWin, IDC_READ_DEFECTS, STR(BUTTON_READ_DEFECT) );
      SetDlgItemText( hAppWin, IDC_REMAP_DEFECTS, STR(BUTTON_REMAP_DEFECT) );
      SetDlgItemText( hAppWin, IDC_OPTIONS, STR(BUTTON_OPTIONS) );
      return true;
   }
   return false;
}

bool SetManualLanguage()
{
   SetWindowText( hManDlg, STR(MANUAL_CONNECTION) );
   SetDlgItemText( hManDlg, IDCANCEL, STR(BUTTON_CANCEL) );
   SetDlgItemText( hManDlg, IDOK, STR(BUTTON_CONNECT) );
   SetDlgItemText( hManDlg, IDC_OPTIONS, STR(BUTTON_OPTIONS) );
   return true;
}

void SetOptionsLanguage()
{
   SetWindowText( hOptDlg, STR(BUTTON_OPTIONS) );
   SetDlgItemText( hOptDlg, IDCANCEL, STR(BUTTON_CLOSE) );
   SetDlgItemText( hOptDlg, IDC_ENABLE_PARSING, STR(CHECKBOX_ENABLE_PARSING) );
   SetDlgItemText( hOptDlg, IDC_LANGUAGE_TEXT, STR(STATIC_LANGUAGE) );
}

enum OSV { OSV_NONE, OSV_USB_IOCTL = 1, OSV_USB_ASPI = 2, OSV_COM = 4 };
unsigned osv_mask;

bool CheckConnection( HWND hDlg )
{
   buffer[0] = '\0';

   int i = SendDlgItemMessage(hDlg, IDC_DEVICE_LIST, CB_GETCURSEL, 0, 0 );
   if( i == CB_ERR )
      return false;

   unsigned param = SendDlgItemMessage(hDlg, IDC_DEVICE_LIST, CB_GETITEMDATA, i, 0);

   if( (param & PARAM_TYPE) == PARAM_USB_IOCTL )
   {
      CheckUSB((char)(param & 0x7F), buffer, sizeof(buffer));
   }
#ifdef SUPPORT_SERIAL_COM
   if( (param & PARAM_TYPE) == PARAM_COM )
   {
      CheckCOM((char)(param & 0x7F), buffer, sizeof(buffer));
   }
#endif
#ifdef SUPPORT_WINDOWS_98SE
   if( (param & PARAM_TYPE) == PARAM_USB_ASPI )
   {
      CheckUSB_ASPI((param >> 16)&255, (param >> 8)&255, (param)&255, buffer, sizeof(buffer));
   }
#endif

   SetDlgItemText(hDlg, IDC_RESULT_TEXT, buffer);
   return true;
}

void RescanDevices(HWND hDlg)
{
   DOUT(("Scan Devices\n"));

   HWND hCbox = GetDlgItem(hDlg, IDC_DEVICE_LIST);
   int i;
   unsigned param;

   i = SendMessage(hCbox, CB_ADDSTRING, 0, (LPARAM)"Select device");
   SendMessage(hCbox, CB_SETITEMDATA, i, (LPARAM)0);

   if( osv_mask & OSV_USB_IOCTL )
   {
      IOCTL_DriveName ioctl;
      const char* drive;
      while( (drive = ioctl.GetNext(param)) != NULL )
      {
         i = SendMessage(hCbox, CB_ADDSTRING, 0, (LPARAM)drive);
         SendMessage(hCbox, CB_SETITEMDATA, i, (LPARAM)param);
      }
   }
#ifdef SUPPORT_WINDOWS_98SE
   if( osv_mask & OSV_USB_ASPI )
   {
      ASPI_HostName host;
      const char* name;
      while( (name = host.GetNext(param)) != NULL )
      {
         i = SendMessage(hCbox, CB_ADDSTRING, 0, (LPARAM)name);
         SendMessage(hCbox, CB_SETITEMDATA, i, (LPARAM)param);
      }
   }
#endif
#ifdef SUPPORT_SERIAL_COM
   if( osv_mask & OSV_COM )
   {
      COM_PortName com;
      const char* port;
      while( (port = com.GetNext(param)) != NULL )
      {
         i = SendMessage(hCbox, CB_ADDSTRING, 0, (LPARAM)port);
         SendMessage(hCbox, CB_SETITEMDATA, i, (LPARAM)param);
      }
   }
#endif

   SendMessage(hCbox, CB_SETCURSEL, 0, 0 );
   PostMessage(hDlg, WM_COMMAND, MAKEWPARAM(IDC_DEVICE_LIST,CBN_SELCHANGE), (LPARAM)hCbox );
}

BOOL ManualOptionsCommand(HWND hDlg, int NotifyCode, int id, HWND hCtrl)
{
   NotifyCode;
   hCtrl;

   switch(id)
   {
   case IDCANCEL:
      PostMessage(hDlg, WM_CLOSE, 0,0);
      return TRUE;

   case IDC_LANGUAGE:
      if( NotifyCode == CBN_SELCHANGE )
      {
         SetLanguage( SendMessage(hCtrl, CB_GETCURSEL, 0, 0 ) );
         SetOptionsLanguage();
         SetManualLanguage();
         return TRUE;
      }
      return FALSE;

   default:
      return FALSE;
   }
}


BOOL CALLBACK ManualOptionsDlg( HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
   if( uMsg == WM_INITDIALOG )
   {
      hOptDlg = hDlg;
      HWND hLangWin = GetDlgItem(hDlg, IDC_LANGUAGE);

      int save_lang = Language::curr_lang;
      for(int i=0; i<Language::LanguageCount; ++i)
      {
         Language::curr_lang = i;
         SendMessage(hLangWin, CB_ADDSTRING, 0, (LPARAM)STR(MESSAGE_LANGUAGE) );
      }
      Language::curr_lang = save_lang;
      SetOptionsLanguage();
      SendMessage(hLangWin, CB_SETCURSEL, Language::curr_lang, 0 );
      return TRUE;
   }

   if( uMsg == WM_COMMAND )
   {
      return ManualOptionsCommand(hDlg, HIWORD(wParam), LOWORD(wParam), (HWND) lParam);
   }

   if( uMsg == WM_CLOSE )
   {
      hOptDlg = NULL;
      EndDialog(hDlg, IDCLOSE);
      return FALSE;
   }
   return FALSE;
}

void ManualOptions( HWND hParent )
{
   int ret = DialogBox( hAppInst, MAKEINTRESOURCE(IDD_MANUAL_OPTIONS), hParent, ManualOptionsDlg);
   if( ret != IDOK ) { DOUT(("ManualOptionsDlg: %d\n", ret)) }
}

BOOL ManualConnectionCommand(HWND hDlg, int NotifyCode, int id, HWND hCtrl)
{
   NotifyCode;
   hCtrl;

   switch(id)
   {
   case IDC_OPTIONS:
      ManualOptions( hDlg );
      return TRUE;

   case IDC_CHECK_CONNECTION:
      CheckConnection( hDlg );
      return TRUE;

   case IDC_RESCAN_DEVICES:
      SendDlgItemMessage(hDlg, IDC_DEVICE_LIST, CB_RESETCONTENT, 0, 0 );
      RescanDevices(hDlg);
      return TRUE;

   case IDC_DEVICE_LIST:
      if( NotifyCode == CBN_SELCHANGE )
      {
         int i = SendDlgItemMessage(hDlg, IDC_DEVICE_LIST, CB_GETCURSEL, 0, 0 );
         if( i == CB_ERR )
            return TRUE;

         if( i == 0 )
         {
            SetDlgItemText(hDlg, IDC_RESULT_TEXT, STR(CHOOSE_AND_PRESS_CONTINUE) );
            UpdateWindow(GetDlgItem(hDlg, IDC_RESULT_TEXT));
         }
         else
         {
            char* cb_str = &buffer[80];
            SendDlgItemMessage(hDlg, IDC_DEVICE_LIST, CB_GETLBTEXT, i, (LPARAM)cb_str );

            wsprintf( buffer, "Check connection to %s\r\n", cb_str );
            SetDlgItemText(hDlg, IDC_RESULT_TEXT, buffer);
            UpdateWindow(GetDlgItem(hDlg, IDC_RESULT_TEXT));

            PostMessage(hDlg, WM_COMMAND, IDC_CHECK_CONNECTION, 0);
         }
         return TRUE;
      }
      return FALSE;

   case IDOK:
      {
         int i = SendDlgItemMessage(hDlg, IDC_DEVICE_LIST, CB_GETCURSEL, 0, 0 );
         if( i == CB_ERR )
            return TRUE;

         unsigned param = SendDlgItemMessage(hDlg, IDC_DEVICE_LIST, CB_GETITEMDATA, i, 0);

         if( (param & PARAM_TYPE) == PARAM_USB_IOCTL )
         {
            dsc = OpenUSB( (char)(param & 0x7F) );
         }
#ifdef SUPPORT_SERIAL_COM
         if( (param & PARAM_TYPE) == PARAM_COM )
         {
            dsc = OpenCOM((char)(param & 0x7F));
         }
#endif
#ifdef SUPPORT_WINDOWS_98SE
         if( (param & PARAM_TYPE) == PARAM_USB_ASPI )
         {
            dsc = OpenUSB_ASPI( (param >> 16)&255, (param >> 8)&255, (param)&255);
         }
#endif

         if( dsc )
         {
            SetDlgItemText(hDlg, IDC_RESULT_TEXT, "OK\r\n");
            EndDialog(hDlg, IDOK);
         }
         else
         {
            SetDlgItemText(hDlg, IDC_RESULT_TEXT, "Connection fault\r\n");
         }
         return TRUE;
      }

   case IDCANCEL:
      EndDialog(hDlg, IDCANCEL);
      return TRUE;

   default:
      return FALSE;
   }
}

BOOL CALLBACK ManualConnectDlg( HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
   if( uMsg == WM_INITDIALOG )
   {
      hManDlg = hDlg;
      RescanDevices(hDlg);
      return TRUE;
   }

   if( uMsg == WM_COMMAND )
   {
      return ManualConnectionCommand(hDlg, HIWORD(wParam), LOWORD(wParam), (HWND) lParam);
   }

   if( uMsg == WM_CLOSE )
   {
      hManDlg = NULL;
      EndDialog(hDlg, IDCLOSE);
      return FALSE;
   }
   return FALSE;
}

int GetOSVersion()
{
   OSVERSIONINFO osv;
   osv.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
   if( !GetVersionEx( &osv ) ) return OSV_NONE;

   DOUT(("GetVersionEx: %u v%u.%02u.%04u (%s)\n", osv.dwPlatformId, osv.dwMajorVersion, osv.dwMinorVersion, osv.dwBuildNumber&0xFFFF, osv.szCSDVersion));

   if( osv.dwMajorVersion == 4 )
      return OSV_USB_ASPI|OSV_COM;

   if( osv.dwMajorVersion >= 5 )
      return OSV_USB_IOCTL|OSV_COM;

   return OSV_COM;
}

bool Connect(HWND hDlg)
{
   DOUT(("Auto connection\n"));

   TDeviceInfo info;

   osv_mask = GetOSVersion();
   if( !dsc && (osv_mask & OSV_USB_IOCTL) )
   {
      dsc = OpenUSB();
   }
#ifdef SUPPORT_WINDOWS_98SE
   if( !dsc && (osv_mask & OSV_USB_ASPI) )
   {
      dsc = OpenUSB_ASPI();
   }
#endif
#ifdef SUPPORT_SERIAL_COM
   if( !dsc && (osv_mask & OSV_COM) )
   {
      dsc = OpenCOM();
   }
#endif

   if( !dsc )
   {
      if( IDYES != MessageBox(hDlg, STR(CANNOT_OPEN_CAMERA), program, MB_YESNO|MB_ICONEXCLAMATION) )
         return false;

      int ret = DialogBox( hAppInst, MAKEINTRESOURCE(IDD_MANUAL_CONNECT), hDlg, ManualConnectDlg);
      if( ret != IDOK ) { DOUT(("ManualConnectDlg: %d\n", ret)) }

      if( !dsc )
         return false;
   }

   int len = 0;
   if( dsc->GetDeviceInfo( info ) )
   {
      len += wsprintf(&buffer[len], "%s%s%s\r\n",
         info.VendorId, info.ProductId, info.ProductRevisionLevel);
   }
   else
   {
      len += wsprintf(&buffer[len], "%s\r\n", dsc->comment );
   }

   TResponce responce;

   if( dsc->GetVar(0x1A, responce) )
   {
      len += wsprintf(&buffer[len], "FIRMWARE: %s\r\n", responce.GetStr() );
   }
   if( dsc->GetVar(0x19, responce) )
   {
      len += wsprintf(&buffer[len], "S/N: %s\r\n", responce.GetStr() );
   }

   command = &buffer[len];

   if( hMsgWin ) SetWindowText(hMsgWin, buffer);

   EnableControls( hDlg, TRUE );
   return true;
}

void InitDlg(HWND hDlg)
{
   hMsgWin = GetDlgItem(hDlg, IDC_TABLE);
   hProgressWin = GetDlgItem(hDlg, IDC_PROGRESS);
   SetLanguage( Language::EN );

   EnableControls( hDlg, FALSE );

   if( hMsgWin ) SetWindowText(hMsgWin, STR(CONNECTING_TO_CAMERA));

   PostMessage(hDlg, WM_COMMAND, IDC_CONNECT, 0);
}

BOOL OptionsCommand(HWND hDlg, int NotifyCode, int id, HWND hCtrl)
{
   NotifyCode;
   hCtrl;

   switch(id)
   {
   case IDCANCEL:
      PostMessage(hDlg, WM_CLOSE, 0,0);
      return TRUE;

   case IDC_LANGUAGE:
      if( NotifyCode == CBN_SELCHANGE )
      {
         SetLanguage( SendMessage(hCtrl, CB_GETCURSEL, 0, 0 ) );
         SetOptionsLanguage();
         return TRUE;
      }
      return FALSE;

   case IDC_ENABLE_PARSING:
      enable_parcing = !enable_parcing;
      CheckDlgButton(hDlg, IDC_ENABLE_PARSING, enable_parcing ? BST_CHECKED : BST_UNCHECKED );
      EnableControls(hAppWin, TRUE);
      return TRUE;
   default:
      return FALSE;
   }
}


BOOL CALLBACK OptionsDlg( HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
   if( uMsg == WM_INITDIALOG )
   {
      hOptDlg = hDlg;
      CheckDlgButton(hDlg, IDC_ENABLE_PARSING, enable_parcing ? BST_CHECKED : BST_UNCHECKED );

      HWND hLangWin = GetDlgItem(hDlg, IDC_LANGUAGE);

      int save_lang = Language::curr_lang;
      for(int i=0; i<Language::LanguageCount; ++i)
      {
         Language::curr_lang = i;
         SendMessage(hLangWin, CB_ADDSTRING, 0, (LPARAM)STR(MESSAGE_LANGUAGE) );
      }
      Language::curr_lang = save_lang;
      SetOptionsLanguage();
      SendMessage(hLangWin, CB_SETCURSEL, Language::curr_lang, 0 );
      return TRUE;
   }

   if( uMsg == WM_COMMAND )
   {
      return OptionsCommand(hDlg, HIWORD(wParam), LOWORD(wParam), (HWND) lParam);
   }

   if( uMsg == WM_CLOSE )
   {
      hOptDlg = NULL;
      EndDialog(hDlg, IDCLOSE);
      return FALSE;
   }
   return FALSE;
}

void Options( HWND hParent )
{
   int ret = DialogBox( hAppInst, MAKEINTRESOURCE(IDD_OPTIONS), hParent, OptionsDlg);
   if( ret != IDOK ) { DOUT(("OptionsDlg: %d\n", ret)) }
}

BOOL Command(HWND hDlg, int NotifyCode, int id, HWND hCtrl)
{
   NotifyCode;
   hCtrl;

   switch(id)
   {
   case IDCANCEL:
      if( show_stop )
         Stop( hDlg );
      else
         EndDialog(hDlg, IDCANCEL);
      return TRUE;

   case IDC_CONNECT:
      if( !Connect( hDlg ) )
         EndDialog(hDlg, IDABORT);
      return TRUE;

   case IDC_OPTIONS: Options( hDlg ); return TRUE;
   case IDC_REMAP_DEFECTS: Remap( hDlg, FALSE ); return TRUE;
   case IDC_READ_DEFECTS: Read( hDlg, FALSE ); return TRUE;
   default:
      return FALSE;
   }
}

void EnableControls(HWND hDlg, BOOL en)
{
   EnableWindow( GetDlgItem(hDlg, IDC_OPTIONS), en );
   EnableWindow( GetDlgItem(hDlg, IDC_REMAP_DEFECTS), en );
   EnableWindow( GetDlgItem(hDlg, IDC_READ_DEFECTS), en );
   EnableWindow( GetDlgItem(hDlg, IDCANCEL), en );
}

BOOL CALLBACK MainDlg( HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
   if( uMsg == WM_INITDIALOG )
   {
      if( !hAppWin )
         hAppWin = hDlg;

      InitDlg(hDlg);
      return TRUE;
   }

   if( uMsg == WM_TIMER )
   {
      if( wParam == IDT_READ )
      {
         if( read_status > READ_IN_PROGRESS )
            read_count = MAX_READ;

         SendMessage(hProgressWin, PBM_SETPOS, read_count, 0 );

         if( read_count == MAX_READ )
         {
            is_busy = false;
            KillTimer( hDlg, IDT_READ );

            if( show_stop )
            {
               show_stop = false;
               SetDlgItemText(hDlg, IDCANCEL, STR(BUTTON_CLOSE));
            }
            EnableControls(hDlg, TRUE);
         }

         SetStatusLine();
         message[0] = '\r';
         message[1] = '\n';

         if( hMsgWin ) SetWindowText(hMsgWin, buffer);

         return TRUE;
      }

      if( wParam == IDT_REMAP )
      {
         ++remap_count;
         if( remap_status > REMAP_IN_PROGRESS )
            remap_count = MAX_REMAP;

         SendMessage(hProgressWin, PBM_SETPOS, remap_count, 0 );

         unsigned sec = (GetTickCount() - start + 500)/1000;
         
         if( remap_count == MAX_REMAP )
         {
            is_busy = false;
            KillTimer( hDlg, IDT_REMAP );
            EnableControls(hDlg, TRUE);

            if( remap_status == REMAP_IN_PROGRESS )
               wsprintf( status, STR(TIME_IS_UP), sec );
            else
            if( remap_status == REMAP_COMPLETE )
               wsprintf( status, STR(REMAP_COMPLETE), sec );
            else
            if( remap_status == REMAP_INTERNAL_ERROR )
               wsprintf( status, STR(REMAP_INTERNAL_ERROR), sec );
            else
            if( remap_status == REMAP_OK )
               wsprintf( status, STR(REMAP_OK0), sec );
            else
            if( remap_status == REMAP_OK+1 )
               wsprintf( status, STR(REMAP_OK1), sec );
            else
            if( remap_status > REMAP_OK )
               wsprintf( status, STR(REMAP_OKX), sec, remap_status-REMAP_OK );
            else
               wsprintf( status, STR(REMAP_FAULT), sec );
         }
         else
         {
            wsprintf( status, STR(ELAPSED_TIME), sec);
         }

         if( hMsgWin ) SetWindowText(hMsgWin, buffer);
         return TRUE;
      }

      return FALSE;
   }
   if( uMsg == WM_COMMAND )
   {
      return Command(hDlg, HIWORD(wParam), LOWORD(wParam), (HWND) lParam);
   }
   if( uMsg == WM_CLOSE )
   {
      EndDialog(hDlg, IDCLOSE);
      return TRUE;
   }

   if( uMsg == WM_DESTROY )
   {
      if( hDlg == hAppWin )
         hAppWin = NULL;

      hMsgWin = hProgressWin = NULL;
      return FALSE;
   }
   
   return FALSE;
}

int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE, LPSTR, int )
{
   hAppInst = hInstance;

   InitCommonControls();

#ifdef DEBUG
   open_log("cpixmap.log");
#endif

   DOUT(("Run: cpixmap " version " [" compilation_time "]\n"))

   int ret = DialogBox( hAppInst, MAKEINTRESOURCE(IDD_MAIN), NULL, MainDlg);

   if( ret != IDOK ) { DOUT(("MainDlg: %d\n", ret)) }

   delete dsc;
   dsc = NULL;

   DOUT(("Normal exit\n"))

#ifdef DEBUG
   close_log();
#endif

   return 0;
}

